/*
 *  Copyright (c) 2017 Viktor Odintsev <ninetls@xfce.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef __SN_UTIL_H__
#define __SN_UTIL_H__

#include <gtk/gtk.h>

G_BEGIN_DECLS

gulong
sn_signal_connect_weak (gpointer instance,
                        const gchar *detailed_signal,
                        GCallback c_handler,
                        gpointer data);

gulong
sn_signal_connect_weak_swapped (gpointer instance,
                                const gchar *detailed_signal,
                                GCallback c_handler,
                                gpointer data);

gboolean
sn_container_has_children (GtkWidget *widget);

G_END_DECLS

#endif /* !__SN_UTIL_H__ */
